/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: TTL.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

/* Common Library */
#include "SOAPMessage.h"
#include "FileIDFunc.h"

/* Server Common Library */
#include "ServerCommon.h"
#include "ChannelDescriptor.h"

#include "TTL.h"

using namespace OpenSOAP;
// Constructor
// 
TTL::TTL(char filePass[100])
{
  // Multimap Key: ExpireTime,
  //        Value: Structure[msgID,fileID,startTime,ttl,backwardPath]
  mapRecord m;

  time_t expireTime;

  ttlMap.clear();
  strcpy(tableFile, filePass);

  // Read from Backup File
  //
  min.open(tableFile, ios::in);
  if (min.fail()) {
    cout << "TTL::Create map file" << endl;
  }
  while (!min.eof() && !min.fail()) {
    min >> expireTime;
    min >> m.msgID;
    min >> m.fileID;
    min >> m.startTime;
    min >> m.ttl;
    min >> m.backwardPath;
    if (!min.eof()) {
      cout << "Expire Datetime = " << asctime(localtime(&expireTime));
      cout << "msgID: " << m.msgID << "\t" 
	   << "fileID: " << m.fileID << "\t" 
	   << "TTL: " << m.ttl << "sec" << endl; 
      cout << "BackwardPath: " << m.backwardPath << endl;
      cout << "Start Datetime = " << asctime(localtime(&m.startTime));
      cout << "----------------" << endl;

      ttlMap.insert(pair<time_t, mapRecord>(expireTime, m));
    }
  }
  min.close();

  // Thread Initialize
  pthread_mutex_init(&ttl_mutex, NULL);
  pthread_create(&watcher_thread, NULL, watcher, this);
}

void
TTL::setChnlDesc(const ChannelDescriptor& chnlDesc)
{
  cerr << "TTL::setChnlDesc" << endl;
  chnlDesc_ = chnlDesc;
}

// Backup File Output
// Return Value [success:0, failure:1]
//
int TTL::fileOut()
{
  mout.open(tableFile, ios::out);
  if (mout.fail()) {
    cout << "TTL::Map_file output open error" << endl;
    return 1;
  }


  multimap<time_t, mapRecord>::iterator itr = ttlMap.begin();
  multimap<time_t, mapRecord>::iterator itrEnd = ttlMap.end();
  for (; itr != itrEnd; itr++) {
    mout << itr->first << "\t" 
	 << itr->second.msgID << "\t" << itr->second.fileID << "\t" 
	 << itr->second.startTime << "\t" << itr->second.ttl << "\t"
	 << itr->second.backwardPath << endl ;
    cout << itr->first << "\t"
	 << itr->second.msgID << "\t" << itr->second.fileID << "\t" 
	 << itr->second.startTime << "\t" << itr->second.ttl << "\t"
	 << itr->second.backwardPath << endl ;
    mout.flush();
  }

  mout.close();
  return 0;
}


// Push Method
// Return Value [success:0, failure: not 0]
//
int TTL::push(string msgID, string fileID, time_t startTime,
	      int ttl, string backwardPath)
{
  mapRecord m;
  
  if (ttl > ttlMaxValue) {
    cout << "TTL::too large ttl value" << endl;
    return 1;
    exit(1);
  }

  m.msgID = msgID;
  m.fileID = fileID;
  m.startTime = startTime;
  m.ttl = ttl;
  m.backwardPath = backwardPath;

  // Mutex Lock
  int sl = pthread_mutex_lock(&(ttl_mutex));
  if (sl != 0) {
    cout << "TTL::push thread mutex lock error" << endl;
    return 1;
  }

  // Insert to Multimap
  ttlMap.insert(pair<time_t, mapRecord>(startTime + ttl, m)); 

  cout << "map size = " << ttlMap.size() << endl ;

  // File Output
  int sf = fileOut();
  if (sf != 0) {
    cout << "TTL::Map file write error" << endl;
    return 2;
  }


  // Mutex Unlock
  int su = pthread_mutex_unlock(&(ttl_mutex));
  if (su != 0) {
    cout << "TTL::push thread mutex unlock error" << endl;
    return 3;
  }
  else if (su == 0) {
    cout << "TTL::push thread mutex unlocked" << endl;
  }

  return 0;
}


// Watcher Thread
//
void* TTL::watcher(void* arg)
{
  // Pointer to Instance
  TTL* that = (TTL*)arg;
  if (NULL == that) {
    cerr << "TTL::Invalid argument" << endl;
  }

  while (1) {
    sleep(that->watchInterval);

    // Serch the Expirable Record in the TTLTable
    multimap<time_t, mapRecord>::iterator p = that->ttlMap.begin();
    multimap<time_t, mapRecord>::iterator pEnd = that->ttlMap.end();

    if (p != pEnd && p->first < time(NULL)) {

      // Mutex Lock
      int sl = pthread_mutex_lock(&(that->ttl_mutex));
      if (sl != 0) {
	cout << "TTL::watcher thread mutex lock error" << endl;
      }

      for (; p != pEnd && p->first < time(NULL); p++) {
	// Find the Expirable Record
	cout << "TimeOut msgID = " << p->second.msgID << endl;

	// Send the Signal to Queue,SpoolManager
	int se = that->eraseSignalToFIFO(p->second.msgID, p->second.fileID);
	if (se != 0) {
	  cout << "TTL::Erase Signal FIFO error" << endl;
	}

	that->ttlMap.erase(p);

	cout << "map size = " << that->ttlMap.size() << endl ;    
      }

      // Backup File Output
      int sf = that->fileOut();
      if (sf != 0) {
	cout << "TTL::Map file write error" << endl;
      }

      // Mutex unlock
      int su = pthread_mutex_unlock(&(that->ttl_mutex));
      if (su == 0) {
	cout << "TTL::watcher thread mutex unlocked" << endl;
      }
    }
    //else if (p != pEnd && p->first > time(NULL)) {
    //  cout << "Not expired yet" << endl;
    //}
  }
}

int TTL::eraseSignalToFIFO(string msgID, string fileID)
{

  // Please Implement!
  // Queue Manager << fileID
  // Spool Manager << msgID

  // cout << "FIFO:msgID = " << msgID << endl;
  // cout << "FIFO:fileID = " << fileID << endl;

  return 0;

}

string TTL::refBackwardPath(string msgID)
{
  string backwardPath = "No Entry";

  // Mutex lock
  int sl = pthread_mutex_lock(&(ttl_mutex));
  if (sl != 0) {
    cout << "TTL::reference method mutex lock error" << endl;
  }

  // Find the BackwardPath correponding to msgID
  multimap<time_t, mapRecord>::iterator p = ttlMap.begin();
  for (; p != ttlMap.end(); p++) {
    if (p->second.msgID == msgID) {
      backwardPath = p->second.backwardPath;
    }
  }

  // Mutex unlock
  int su = pthread_mutex_unlock(&(ttl_mutex));
  if (su != 0) {
    cout << "TTL::reference method mutex unlock error" << endl;
  }

  return backwardPath;

}

// Destructor
//
TTL::~TTL()
{
  pthread_join(watcher_thread, NULL);
}

bool 
TTL::execPushOpr()
{
  cerr << "TTL::execPushOpr()" << endl;
  
  // MsgDrv 瑗Ăf[^ (t@C ID) ǂݍށB
  string fileIDOfMsgFromMsgDrv;
  if (0 > chnlDesc_.read(fileIDOfMsgFromMsgDrv)) {
    cerr << TAG_ERR << "TTL::execPushOpr: read failed" << endl;
    return false;
  }
  
  cerr << "** TTL ** READ From MsgDrv=[" 
       << fileIDOfMsgFromMsgDrv << "]" << endl;
  std::string soapMsgFromMsgDrv = revertFromFileID(fileIDOfMsgFromMsgDrv);
  
  std::string msgIDContent = getMsgIDContent(soapMsgFromMsgDrv);
  
  std::string backwardPath;
  getBackwardPath(soapMsgFromMsgDrv, backwardPath);
  bool ttlHeaderExist;
  unsigned int ttl = 80;
  std::string ttlType;
  getTTLInfo(soapMsgFromMsgDrv, ttlHeaderExist, ttl, ttlType);
  
  // TTL Table ɓB
  if (msgIDContent != "") {
    int sp = push(msgIDContent, fileIDOfMsgFromMsgDrv, 
		  time(NULL), ttl, backwardPath);
    if (sp != 0) {
      cout << "Map insert error" << endl;
    }
  }
  
  // TTL Table ɓƂ̊mFbZ[W MsgDrv ɕԂB
  string fileIDOfMsgToMsgDrv = "TTL Table Push success";
  if (0 > chnlDesc_.write(fileIDOfMsgToMsgDrv)) {
    cerr << TAG_ERR << "TTL::execPushOpr: write failed" << endl;
    return false;
  }
  
  cerr << "** Spool ** WRITE To MsgDrv=[" 
       << fileIDOfMsgToMsgDrv << "]" << endl;
  
  return true;
}// end of execPushOpr()


bool 
TTL::execRefOpr()
{
  
  // Forwarder 瑗Ăf[^ (t@C ID) ǂݍށB
  std::string msgIDContent;
  if (0 > chnlDesc_.read(msgIDContent)) {
    cerr << TAG_ERR << "TTL::execPushOpr: read failed" << endl;
    return false;
  }
  
  cerr << "** TTL ** READ From Fwdr=[" 
       << msgIDContent << "]" << endl;
  
  std::string backwardPath = refBackwardPath(msgIDContent);
  cout << "BackwardPath= " << backwardPath << endl;
  
  // Forwarder ɊY Msg ID X|XԂB
  
  if (0 > chnlDesc_.write(backwardPath)) {
    cerr << TAG_ERR << "TTL::execPopOpr: write failed" << endl;
    return false;
  }
  
  cerr << "** TTL ** WRITE To Forwarder=[" 
       << backwardPath << "]" << endl;
  
  return true;
}// end of execRefOpr()

/*
int main()
{
  string msgID;
  string fileID;
  string backwardPath;
  int ttl;
  int f, sp;

  time_t now = time(NULL);
  cout << "TTL::start  " << asctime(localtime(&now));

  // Generate an Instance
  TTL TTLMgr("./ttl_map");

  TTLMgr.watchInterval = 1;
  TTLMgr.ttlMaxValue = 3600;

  // Test Loop
  while (1) {
    sleep(1);
    cout << "push(1) or reference BackwardPath(2) ? ";
    cin >> f;

    switch (f) {
    case 1:
      cout << "push process" << endl;
      cout << "msg ID ? ";
      cin >> msgID;
      cout << "file ID ? ";
      cin >> fileID;
      cout << "TTL(second) ? ";
      cin >> ttl;
      cout << "Backward Path ? ";
      cin >> backwardPath;
      sp = TTLMgr.push(msgID, fileID, time(NULL), ttl, backwardPath);
      if (sp != 0) {
	cout << "Map insert error" << endl;
      }
      break;

    case 2:
      cout << "reference BackwardPath process" << endl;
      cout << "msgID ? ";
      cin >> msgID;
      backwardPath = TTLMgr.refBackwardPath(msgID);
      cout << "BackwardPath= " << backwardPath << endl;
      break;

    default:
      exit(1);
    }
  }
  
}

*/

